from enum import Enum
from typing import Any, List, Callable

from grafana_tools.domain.model import (
    Account, DirectoryService, Grafana, Logger, UserNotFound, Organization
)


class Actions(Enum):
    NOTHING = 1
    NOTIFY = 2
    REMOVE = 3


def synchronize(
    grafana: Grafana, information: DirectoryService, **kwargs: Any
) -> None:
    '''Update Grafana based on a Directory Information Service'''

    def info(message: str) -> None:
        if (
            'logger' in kwargs and
            isinstance(kwargs['logger'], Logger)
        ):
            kwargs['logger'].info(message)

    info('Starting synchonization')

    ee_employees = information.get_ee_employees()
    ensure_corresponding_users_exist(ee_employees, grafana, info)

    ee_admin_logins = [e.uid for e in information.get_ee_admins()]
    ee_non_admin_logins = [
        e.uid
        for e in ee_employees
        if e.uid not in ee_admin_logins
    ]
    ee_all_logins = add_lists(ee_admin_logins, ee_non_admin_logins)

    for organization in grafana.list_organizations():
        info('Organization "{}/{}"'.format(organization.id, organization.name))

        if organization.is_client():
            accounts = information.metrology_accounts_for(organization.trigram)
            ensure_corresponding_users_exist(accounts, grafana, info)

            logins = [a.uid for a in accounts]
            grafana.set_role_for_all(
                logins, grafana.VIEWER, organization
            )
            handle_expired_users(
                grafana,
                add_lists(logins, ee_all_logins),
                organization,
                kwargs.get('expired_users', Actions.NOTHING),
                info
            )

        if organization.is_client() or organization.is_easter_eggs():
            grafana.set_role_for_all(
                ee_admin_logins, grafana.ADMINISTRATOR, organization
            )

            grafana.set_role_for_all(
                ee_non_admin_logins, grafana.EDITOR, organization
            )


def add_lists(*args: List[str]) -> List[str]:
    logins = []
    for a in args:
        logins.extend(a)
    return logins


def ensure_corresponding_users_exist(
    some_accounts: List[Account], grafana: Grafana, log: Callable[[str], None]
) -> None:
    log('Making sure all accounts have a corresponding user in Grafana')

    for account in some_accounts:
        try:
            user = grafana.find_user(account.uid)
            log('Update user "{}"'.format(account.uid))
            grafana.update_user(user.id, account.uid, account.cn, account.mail)
        except UserNotFound:
            log('Create user "{}"'.format(account.uid))
            grafana.create_user(account.uid, account.cn, account.mail)


def handle_expired_users(
    grafana: Grafana,
    logins_to_keep: List[str],
    organization: Organization,
    action: Actions,
    log: Callable[[str], None]
) -> None:
    if action == Actions.NOTHING:
        return

    log('Analyzing expired users for {}'.format(organization))

    for user in grafana.users_in(organization):
        if user.login in logins_to_keep:
            continue
        if action == Actions.NOTIFY:
            log('{} not in organization anymore'.format(user))
        elif action == Actions.REMOVE:
            log('Removing {}'.format(user))
            grafana.remove_user_from(user, organization)
