from typing import List, Dict, Tuple


class Logger:
    def critical(self, message: str) -> None:
        raise NotImplementedError

    def error(self, message: str) -> None:
        raise NotImplementedError

    def warning(self, message: str) -> None:
        raise NotImplementedError

    def info(self, message: str) -> None:
        raise NotImplementedError

    def debug(self, message: str) -> None:
        raise NotImplementedError


class Account:

    def __init__(
        self, uid: str, cn: str, given_name: str, sn: str, mail: str
    ) -> None:
        self.uid = uid
        self.cn = cn
        self.given_name = given_name
        self.sn = sn
        self.mail = mail

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Account):
            return NotImplemented
        return (
            self.uid == other.uid and
            self.cn == other.cn and
            self.given_name == other.given_name and
            self.sn == other.sn and
            self.mail == other.mail
        )

    def __repr__(self) -> str:
        return '<Account "{}">'.format(self.uid)

    def __str__(self) -> str:
        return self.__repr__()


class Credentials:

    def __init__(self, user: str, password: str = '') -> None:
        self.user = user
        self.password = password


class OrganizationException(Exception):
    pass


class NoTrigram(OrganizationException):

    def __init__(self, an_organization: 'Organization') -> None:
        super().__init__(
            'This organization ({}) does not have a trigram'.format(
                an_organization
            )
        )


class Organization:

    def __init__(self, id: int, name: str) -> None:
        self.id = id
        self.name = name

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Organization):
            return NotImplemented
        return (
            self.id == other.id and
            self.name == other.name
        )

    def __repr__(self) -> str:
        return '<Organization "{}" "{}">'.format(
            self.id, self.name
        )

    def __str__(self) -> str:
        return self.__repr__()

    @property
    def trigram(self) -> str:
        if self.is_easter_eggs():
            return 'ee'
        elif self.is_client():
            return self.name[len('cst_'):]
        raise NoTrigram(self)

    def is_client(self) -> bool:
        return self.name.startswith('cst_')

    def is_easter_eggs(self) -> bool:
        return self.name == 'easter-eggs'


class User:

    def __init__(
        self, id: int, login: str, name: str = '', email: str = ''
    ) -> None:
        self.id = id
        self.login = login
        self.name = name
        self.email = email

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, User):
            return NotImplemented
        return (
            self.id == other.id and
            self.login == other.login
        )

    def __repr__(self) -> str:
        return '<User "{}" "{}">'.format(self.id, self.login)

    def __str__(self) -> str:
        return self.__repr__()


class GrafanaException(Exception):
    pass


class UserNotFound(GrafanaException):

    def __init__(self, a_login: str) -> None:
        super().__init__(
            'Cannot find user with login "{}"'.format(a_login)
        )


class CannotCreateUser(GrafanaException):

    def __init__(self, a_login: str) -> None:
        super().__init__(
            'Cannot create user with login "{}"'.format(a_login)
        )


class CannotUpdateUser(GrafanaException):

    def __init__(self, a_login: str) -> None:
        super().__init__(
            'Cannot update user with login "{}"'.format(a_login)
        )


class CannotRemoveUserFromOrganization(GrafanaException):

    def __init__(
        self, a_user: User, an_organization: Organization
    ) -> None:
        super().__init__(
            'Cannot remove user "{}" from {}'.format(
                a_user, an_organization
            )
        )


class CannotUpdateRole(GrafanaException):

    def __init__(
        self, a_user: User, a_role: str, an_organization: Organization
    ) -> None:
        super().__init__(
            'Cannot add role "{}" to {} in {}'.format(
                a_role, a_user, an_organization
            )
        )


class CannotAddRole(GrafanaException):

    def __init__(
        self, a_user: User, a_role: str, an_organization: Organization
    ) -> None:
        super().__init__(
            'Cannot add role "{}" to {} in {}'.format(
                a_role, a_user, an_organization
            )
        )


class OrganizationsNotFound(GrafanaException):

    def __init__(self, message: str = ''):
        super().__init__('Cannot find organizations! {}'.format(message))


class CannotListUsers(GrafanaException):

    def __init__(self, an_organization: Organization) -> None:
        super().__init__(
            'Cannot list users in {}'.format(an_organization)
        )


class Grafana:

    VIEWER = 'Viewer'
    EDITOR = 'Editor'
    ADMINISTRATOR = 'Admin'

    def list_organizations(self) -> List['Organization']:
        raise NotImplementedError

    def set_role_for_all(
        self, logins: List[str], role: str, organization: 'Organization'
    ) -> None:
        raise NotImplementedError

    def find_user(self, login: str) -> 'User':
        raise NotImplementedError

    def set_role(
        self, user: 'User', role: str, organization: 'Organization'
    ) -> None:
        raise NotImplementedError

    def create_user(self, login: str, name: str, email: str) -> 'User':
        raise NotImplementedError

    def update_user(
        self, id: int, login: str, name: str = '', email: str = ''
    ) -> 'User':
        raise NotImplementedError

    def users_in(self, organization: 'Organization') -> List['User']:
        raise NotImplementedError

    def remove_user_from(
            self, user: 'User', organization: 'Organization'
    ) -> None:
        raise NotImplementedError


class DirectoryService:

    def search(self, filters: str) -> List[Tuple[str, Dict[str, List[bytes]]]]:
        raise NotImplementedError

    def get_ee_admins(self) -> List[Account]:
        raise NotImplementedError

    def get_ee_employees(self) -> List[Account]:
        raise NotImplementedError

    def metrology_accounts_for(self, trigram: str) -> List[Account]:
        raise NotImplementedError
