import random
import string
import logging
from typing import Optional, List, Any

import ldap3  # type: ignore
import requests
from requests.auth import HTTPBasicAuth

from grafana_tools.domain.model import (
    Account, Credentials, Organization, User, DirectoryService, Grafana,
    CannotAddRole, CannotCreateUser, CannotListUsers, CannotUpdateRole,
    CannotRemoveUserFromOrganization, UserNotFound, OrganizationsNotFound,
    CannotUpdateUser, Logger,
)


TIMEOUT = 1


class LoggingLogger(Logger):
    def __init__(self, name: str = '') -> None:
        self.__logger = logging.getLogger(name)

    def critical(self, message: str) -> None:
        self.__logger.critical(message)

    def error(self, message: str) -> None:
        self.__logger.error(message)

    def warning(self, message: str) -> None:
        self.__logger.warning(message)

    def info(self, message: str) -> None:
        self.__logger.info(message)

    def debug(self, message: str) -> None:
        self.__logger.debug(message)


class LoggerMixin:

    def log_critical(self, message: str) -> None:
        self.__log('critical', message)

    def log_error(self, message: str) -> None:
        self.__log('error', message)

    def log_warning(self, message: str) -> None:
        self.__log('warning', message)

    def log_info(self, message: str) -> None:
        self.__log('info', message)

    def log_debug(self, message: str) -> None:
        self.__log('debug', message)

    def __log(self, level: str, message: str) -> None:
        if (
            hasattr(self, '_logger') and
            isinstance(getattr(self, '_logger'), Logger) and
            hasattr(getattr(self, '_logger'), level)
        ):
            getattr(getattr(self, '_logger'), level)(message)


class RestApiGrafana(Grafana, LoggerMixin):

    def __init__(
        self,
        url: str,
        credentials: Optional[Credentials] = None,
        **kwargs: Any
    ) -> None:
        self.__url = url
        self.__auth = self.__auth_from_credentials(credentials)
        self._logger = kwargs.get('logger', None)

    def __auth_from_credentials(
        self, credentials: Optional[Credentials] = None
    ) -> Optional[HTTPBasicAuth]:
        if credentials:
            return HTTPBasicAuth(
                credentials.user, credentials.password
            )
        return None

    def list_organizations(self) -> List[Organization]:
        self.log_debug('Looking for organizations')

        response = requests.get(self.__url+'/orgs', auth=self.__auth)

        self.__assert_response_is_ok(response, OrganizationsNotFound())

        return [
            Organization(int(o['id']), o['name'])
            for o in response.json()
        ]

    def set_role_for_all(
        self, logins: List[str], role: str, organization: Organization
    ) -> None:
        self.log_info('Setting role "{}" for: {}'.format(role, logins))

        for login in logins:
            try:
                user = self.find_user(login)
            except UserNotFound as exc:
                self.log_warning(str(exc))
                continue
            self.set_role(user, role, organization)

    def find_user(self, login: str) -> User:
        self.log_debug('Looking for user "{}"'.format(login))

        url = self.__url+'/users?query={}'.format(login)
        response = requests.get(url, auth=self.__auth)

        self.__assert_response_is_ok(response, UserNotFound(login))

        for u in response.json():
            if u.get('login', '') != login:
                continue
            return User(int(u['id']), login)

        raise UserNotFound(login)

    def create_user(self, login: str, name: str, email: str) -> User:
        self.log_debug('Creating user "{}"'.format(login))

        url = self.__url+'/admin/users'
        data = {
            'name': name,
            'email': email,
            'login': login,
            'password': self.__generate_password(),
        }
        response = requests.post(url, auth=self.__auth, json=data)

        self.__assert_response_is_ok(response, CannotCreateUser(login))

        info = response.json()
        return User(int(info['id']), login, name, email)

    def update_user(
        self, id: int, login: str, name: str = '', email: str = ''
    ) -> User:
        self.log_debug('Updating user "{}"'.format(login))

        url = self.__url+'/users/'+str(id)
        data = {
            'login': login,
            'name': name,
            'email': email,
        }
        response = requests.put(url, auth=self.__auth, json=data)

        if response.status_code != 200:
            self.__log_response(response)
            raise CannotUpdateUser(login)

        return User(id, login, name, email)

    def __assert_response_is_ok(
        self, a_response: requests.Response, an_exception: Exception
    ) -> None:
        if not (
            a_response.status_code == 200 and
            a_response.headers.get('content-type', '').startswith(
                'application/json'
            )
        ):
            self.__log_response(a_response)
            raise an_exception

    def __log_response(self, a_response: requests.Response) -> None:
        self.log_debug(
            'Response is {} {}'.format(
                a_response.status_code,
                a_response.headers.get('content-type', 'unknown type')
            )
        )
        self.log_debug(a_response.text)

    def __generate_password(self) -> str:
        return ''.join(
            random.choice(string.ascii_uppercase + string.digits)
            for _ in range(12)
        )

    def set_role(
        self, user: User, role: str, organization: Organization
    ) -> None:
        if user in self.users_in(organization):
            self.__update_role(user, role, organization)
        else:
            self.__add_role(user, role, organization)

    def users_in(self, an_organization: Organization) -> List[User]:
        url = self.__url+'/orgs/{}/users'.format(an_organization.id)
        response = requests.get(url, auth=self.__auth)

        self.__assert_response_is_ok(
            response, CannotListUsers(an_organization)
        )

        return [User(int(u['userId']), u['login']) for u in response.json()]

    def __update_role(
        self, user: User, role: str, organization: Organization
    ) -> None:
        url = self.__url+'/orgs/{}/users/{}'.format(
            organization.id, user.id
        )
        data = {'role': role}
        response = requests.patch(url, auth=self.__auth, json=data)

        self.__assert_response_is_ok(
            response, CannotUpdateRole(user, role, organization)
        )

    def __add_role(
        self, user: User, role: str, organization: Organization
    ) -> None:
        url = self.__url+'/orgs/{}/users'.format(organization.id)
        data = {'loginOrEmail': user.login, 'role': role}
        response = requests.post(url, auth=self.__auth, json=data)

        self.__assert_response_is_ok(
            response, CannotAddRole(user, role, organization)
        )

    def remove_user_from(
            self, a_user: 'User', an_organization: 'Organization'
    ) -> None:
        url = self.__url+'/orgs/{}/users/{}'.format(
            an_organization.id, a_user.id
        )
        response = requests.delete(url, auth=self.__auth)

        self.__assert_response_is_ok(
            response, CannotRemoveUserFromOrganization(a_user, an_organization)
        )


class Ldap(DirectoryService, LoggerMixin):

    def __init__(self, a_server_string: str, **kwargs: Any) -> None:
        ldap3.set_config_parameter('POOLING_LOOP_TIMEOUT', TIMEOUT)
        ldap3.set_config_parameter('RESPONSE_WAITING_TIMEOUT', TIMEOUT)
        ldap3.set_config_parameter('CHECK_AVAILABILITY_TIMEOUT', TIMEOUT)
        try:
            # Not available on older version of ldap3
            ldap3.set_config_parameter('RESET_AVAILABILITY_TIMEOUT', TIMEOUT)
        except ldap3.core.exceptions.LDAPConfigurationParameterError:
            pass

        start_tls = kwargs.get('start_tls', False)
        pool = self.__build_pool(a_server_string, start_tls)

        self.__base = 'o=easter-eggs'
        self.__attrs = ['uid', 'cn', 'givenName', 'sn', 'mail', 'o']
        self.__conn = ldap3.Connection(
            pool,
            auto_bind=True,
            user=kwargs.get('user', None),
            password=kwargs.get('password', None),
        )
        self._logger = kwargs.get('logger', None)

        if start_tls:
            self.__conn.start_tls()

    def __build_pool(
        self, a_server_string: str, start_tls: bool = False
    ) -> List[ldap3.Server]:
        pool = []
        for url in a_server_string.split(','):
            use_ssl = False
            clean_url = url.strip()

            if clean_url.startswith('ldaps://'):
                if start_tls:
                    raise Exception('Cannot use both ldaps:// and START_TLS!')
                use_ssl = True

            server = ldap3.Server(clean_url, use_ssl=use_ssl)
            if server.check_availability():
                pool.append(server)

        if not pool:
            raise Exception('No available server!')

        return pool

    def search(
        self, filters: str
    ) -> List[ldap3.abstract.entry.Entry]:
        if not self.__conn.search(
            self.__base, filters, attributes=self.__attrs
        ):
            return []
        return [e for e in self.__conn.entries]

    def get_ee_admins(self) -> List[Account]:
        return self.__get_accounts_matching(
            '(&(ou=admin)(objectclass=eeuser))'
        )

    def __get_accounts_matching(self, filters: str) -> List[Account]:
        self.log_debug('Looking for accounts matching: "{}"'.format(filters))

        try:
            result = self.search(filters)
        except Exception as exc:
            self.log_warning(
                'An error occured when searching LDAP for "{}"'.format(filters)
            )
            self.log_debug(str(exc))
            return []

        return [
            Account(
                str(info.uid),
                str(info.cn),
                str(info.givenName),
                str(info.sn),
                str(info.mail),
            )
            for info in result
        ]

    def get_ee_employees(self) -> List[Account]:
        return self.__get_accounts_matching(
            '(&'
            '  (|(ou=dev)(ou=com)(ou=admin))'
            '  (objectclass=eeuser)'
            ')'
        )

    def metrology_accounts_for(self, trigram: str) -> List[Account]:
        return self.__get_accounts_matching((
            '(&'
            '  (o={})'
            '  (&'
            '    (objectclass=eecontact)'
            '    (eeallowedservices=outsourcing::metrology)'
            '  )'
            ')'
        ).format(trigram))
