import os
import sys
import logging
import argparse
from typing import Optional

import grafana_tools
from grafana_tools.domain.model import Credentials, Grafana, DirectoryService
from grafana_tools.application import synchronize, Actions
from grafana_tools.infrastructure import RestApiGrafana, Ldap, LoggingLogger


def main() -> None:
    args = parse_args()

    if args.version:
        print_version_and_exit()

    define_logging_level(args.quiet, args.verbose)

    synchronize(
        build_grafana(),
        build_ldap(args),
        expired_users=what_to_do_with_expired_users(args),
        logger=LoggingLogger('synchronize')
    )


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument('--quiet', '-q', action='store_true', default=False)
    parser.add_argument('--verbose', '-v', action='count', default=0)
    parser.add_argument('--version', '-V', action='store_true')
    parser.add_argument('--notify-expired-users', '-e', action='store_true')
    parser.add_argument('--remove-expired-users', '-E', action='store_true')
    parser.add_argument('--start-tls', '-T', action='store_true')

    return parser.parse_args()


def print_version_and_exit() -> None:
    print(grafana_tools.__version__)
    sys.exit(0)


def define_logging_level(quiet: bool, level: int) -> None:
    if quiet:
        logging.basicConfig(level=logging.CRITICAL)
    elif level > 1:
        logging.basicConfig(level=logging.DEBUG)
    elif level > 0:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)


def build_grafana() -> Grafana:
    return RestApiGrafana(
        env('GRAFANA_URL'),
        Credentials(env('GRAFANA_USER'), env('GRAFANA_PASSWORD')),
        logger=LoggingLogger('grafana')
    )


def env(a_name: str, default: Optional[str] = None) -> str:
    if a_name not in os.environ and default is None:
        print('Environment variable "{}" must be defined!'.format(a_name))
        sys.exit(1)
    return os.environ.get(a_name, default or '')


def build_ldap(args: argparse.Namespace) -> DirectoryService:
    try:
        return Ldap(
            env('LDAP_URL'),
            user=env('LDAP_USER', ''),
            password=env('LDAP_PASSWORD', ''),
            logger=LoggingLogger('ldap'),
            start_tls=args.start_tls,
        )
    except Exception as e:
        print('Error instanciating LDAP connection:', e)
        sys.exit(1)


def what_to_do_with_expired_users(args: argparse.Namespace) -> Actions:
    if args.remove_expired_users:
        return Actions.REMOVE
    elif args.notify_expired_users:
        return Actions.NOTIFY
    else:
        return Actions.NOTHING
